@@ -1,19 +1,21 @@
+from __future__ import annotations
+
 import re
 
 from pyinfra.api import FactBase, ShortFactBase
 
 
-class Cpus(FactBase):
+class Cpus(FactBase[int]):
     """
     Returns the number of CPUs on this server.
     """
 
-    command = "getconf NPROCESSORS_ONLN 2> /dev/null || getconf _NPROCESSORS_ONLN"
+    def command(self) -> str:
+        return "getconf NPROCESSORS_ONLN 2> /dev/null || getconf _NPROCESSORS_ONLN"
 
-    @staticmethod
-    def process(output):
+    def process(self, output):
         try:
-            return int(output[0])
+            return int(list(output)[0])
         except ValueError:
             pass
 
@@ -23,11 +25,13 @@ class Memory(FactBase):
     Returns the memory installed in this server, in MB.
     """
 
-    command = "vmstat -s"
-    requires_command = "vmstat"
+    def requires_command(self) -> str:
+        return "vmstat"
 
-    @staticmethod
-    def process(output):
+    def command(self) -> str:
+        return "vmstat -s"
+
+    def process(self, output):
         data = {}
 
         for line in output:
@@ -73,10 +77,12 @@ class BlockDevices(FactBase):
         }
     """
 
-    command = "df"
     regex = r"([a-zA-Z0-9\/\-_]+)\s+([0-9]+)\s+([0-9]+)\s+([0-9]+)\s+([0-9]{1,3})%\s+([a-zA-Z\/0-9\-_]+)"  # noqa: E501
     default = dict
 
+    def command(self) -> str:
+        return "df"
+
     def process(self, output):
         devices = {}
 
@@ -168,9 +174,11 @@ class NetworkDevices(FactBase):
         }
     """
 
-    command = "ip addr show 2> /dev/null || ifconfig -a"
     default = dict
 
+    def command(self) -> str:
+        return "ip addr show 2> /dev/null || ifconfig -a"
+
     # Definition of valid interface names for Linux:
     # https://git.kernel.org/pub/scm/linux/kernel/git/stable/linux.git/tree/net/core/dev.c?h=v5.1.3#n1020
     def process(self, output):
@@ -193,28 +201,36 @@ class NetworkDevices(FactBase):
         output = "\n".join(map(str.strip, output))
 
         # Splitting the output into sections per network device
-        device_sections = re.split(r"\n(?=\d+: \w|\w+:.*mtu.*)", output)
+        device_sections = re.split(r"\n(?=\d+: [^\s/:]|[^\s/:]+:.*mtu )", output)
 
         # Dictionary to hold all device information
         all_devices = {}
 
         for section in device_sections:
             # Extracting the device name
-            device_name_match = re.match(r"^(?:\d+: )?([\w@]+):", section)
+            device_name_match = re.match(r"^(?:\d+: )?([^\s/:]+):", section)
             if not device_name_match:
                 continue
             device_name = device_name_match.group(1)
 
             # Regular expressions to match different parts of the output
-            ether_re = re.compile(r"([0-9A-Fa-f:]{17})")
+            ether_re = re.compile(r"ether ([0-9A-Fa-f:]{17})")
             mtu_re = re.compile(r"mtu (\d+)")
             ipv4_re = (
+                # ip a
                 re.compile(
-                    r"inet (\d+\.\d+\.\d+\.\d+)/(\d+)(?: brd (\d+\.\d+\.\d+\.\d+))"
-                ),  # ip a output,
+                    r"inet (?P<address>\d+\.\d+\.\d+\.\d+)/(?P<mask>\d+)(?: metric \d+)?(?: brd (?P<broadcast>\d+\.\d+\.\d+\.\d+))?"  # noqa: E501
+                ),
+                # ifconfig -a
                 re.compile(
-                    r"inet (\d+\.\d+\.\d+\.\d+)\s+netmask\s+((?:\d+\.\d+\.\d+\.\d+)|(?:[0-9a-fA-FxX]+))(?:\s+broadcast\s+(\d+\.\d+\.\d+\.\d+))"  # noqa: E501
-                ),  # ifconfig -a output
+                    r"inet (?P<address>\d+\.\d+\.\d+\.\d+)\s+netmask\s+(?P<mask>(?:\d+\.\d+\.\d+\.\d+)|(?:[0-9a-fA-FxX]+))(?:\s+broadcast\s+(?P<broadcast>\d+\.\d+\.\d+\.\d+))?"  # noqa: E501
+                ),
+            )
+            ipv6_re = (
+                # ip a
+                re.compile(r"inet6\s+(?P<address>[0-9a-fA-F:]+)/(?P<mask>\d+)"),
+                # ifconfig -a
+                re.compile(r"inet6\s+(?P<address>[0-9a-fA-F:]+)\s+prefixlen\s+(?P<mask>\d+)"),
             )
 
             # Parsing the output
@@ -233,18 +249,22 @@ class NetworkDevices(FactBase):
             )
 
             # IPv4 Addresses
+            ipv4_matches: list[re.Match[str]]
             for ipv4_re_ in ipv4_re:
-                ipv4_matches = ipv4_re_.findall(section)
-                if ipv4_matches:
+                ipv4_matches = list(ipv4_re_.finditer(section))
+                if len(ipv4_matches):
                     break
 
-            if ipv4_matches:
+            if len(ipv4_matches):
                 ipv4_info = []
                 for ipv4 in ipv4_matches:
-                    address = ipv4[0]
-                    mask_value = ipv4[1]
+                    address = ipv4.group("address")
+                    mask_value = ipv4.group("mask")
                     mask_bits, netmask = mask(mask_value)
-                    broadcast = ipv4[2] if len(ipv4) == 3 else None
+                    try:
+                        broadcast = ipv4.group("broadcast")
+                    except IndexError:
+                        broadcast = None
 
                     ipv4_info.append(
                         {
@@ -256,28 +276,24 @@ class NetworkDevices(FactBase):
                     )
                 device_info["ipv4"] = ipv4_info[0]
                 if len(ipv4_matches) > 1:
-                    device_info["ipv4"]["additional_ips"] = ipv4_info[1:]
+                    device_info["ipv4"]["additional_ips"] = ipv4_info[1:]  # type: ignore[index]
 
             # IPv6 Addresses
-            ipv6_re = (
-                re.compile(r"inet6\s+([0-9a-fA-F:]+)/(\d+)"),
-                re.compile(r"inet6\s+([0-9a-fA-F:]+)\s+prefixlen\s+(\d+)"),
-            )
-
+            ipv6_matches: list[re.Match[str]]
             for ipv6_re_ in ipv6_re:
-                ipv6_matches = ipv6_re_.findall(section)
+                ipv6_matches = list(ipv6_re_.finditer(section))
                 if ipv6_matches:
                     break
 
-            if ipv6_matches:
+            if len(ipv6_matches):
                 ipv6_info = []
                 for ipv6 in ipv6_matches:
-                    address = ipv6[0]
-                    mask_bits = ipv6[1] or ipv6[2]
+                    address = ipv6.group("address")
+                    mask_bits = ipv6.group("mask")
                     ipv6_info.append({"address": address, "mask_bits": int(mask_bits)})
                 device_info["ipv6"] = ipv6_info[0]
                 if len(ipv6_matches) > 1:
-                    device_info["ipv6"]["additional_ips"] = ipv6_info[1:]
+                    device_info["ipv6"]["additional_ips"] = ipv6_info[1:]  # type: ignore[index]
 
             all_devices[device_name] = device_info
 
