@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import json
 
 from pyinfra.api import FactBase
@@ -8,11 +10,15 @@ class LxdContainers(FactBase):
     Returns a list of running LXD containers
     """
 
-    command = "lxc list --format json --fast"
-    requires_command = "lxc"
+    def command(self) -> str:
+        return "lxc list --format json --fast"
+
+    def requires_command(self) -> str:
+        return "lxc"
 
     default = list
 
     def process(self, output):
+        output = list(output)
         assert len(output) == 1
         return json.loads(output[0])
