@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import re
 
 from pyinfra.api import FactBase
@@ -9,7 +11,6 @@ class OpenrcStatus(FactBase):
     """
 
     default = dict
-    requires_command = "rc-status"
     regex = (
         r"\s+([a-zA-Z0-9\-_]+)"
         r"\s+\[\s+"
@@ -19,6 +20,9 @@ class OpenrcStatus(FactBase):
         r"\s+\]"
     )
 
+    def requires_command(self, runlevel="default") -> str:
+        return "rc-status"
+
     def command(self, runlevel="default"):
         return "rc-status {0}".format(runlevel)
 
@@ -39,7 +43,9 @@ class OpenrcEnabled(FactBase):
     """
 
     default = dict
-    requires_command = "rc-update"
+
+    def requires_command(self, runlevel="default") -> str:
+        return "rc-update"
 
     def command(self, runlevel="default"):
         self.runlevel = runlevel
