@@ -1,3 +1,7 @@
+from __future__ import annotations
+
+import shlex
+
 from pyinfra.api import FactBase
 
 from .util.packaging import parse_packages
@@ -17,13 +21,14 @@ class PacmanUnpackGroup(FactBase):
         ]
     """
 
-    requires_command = "pacman"
+    def requires_command(self, *args, **kwargs) -> str:
+        return "pacman"
 
     default = list
 
-    def command(self, name):
+    def command(self, package):
         # Accept failure here (|| true) for invalid/unknown packages
-        return 'pacman -S --print-format "%n" {0} || true'.format(name)
+        return 'pacman -S --print-format "%n" {0} || true'.format(shlex.quote(package))
 
     def process(self, output):
         return output
@@ -40,8 +45,11 @@ class PacmanPackages(FactBase):
         }
     """
 
-    command = "pacman -Q"
-    requires_command = "pacman"
+    def command(self) -> str:
+        return "pacman -Q"
+
+    def requires_command(self, *args, **kwargs) -> str:
+        return "pacman"
 
     default = dict
 
