@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 from pyinfra.api import FactBase
 
 from .util.packaging import parse_packages
@@ -16,8 +18,11 @@ class PkginPackages(FactBase):
         }
     """
 
-    command = "pkgin list"
-    requires_command = "pkgin"
+    def command(self) -> str:
+        return "pkgin list"
+
+    def requires_command(self) -> str:
+        return "pkgin"
 
     default = dict
 
