@@ -2,6 +2,8 @@
 Manage ``choco`` (Chocolatey) packages (https://chocolatey.org).
 """
 
+from __future__ import annotations
+
 from pyinfra import host
 from pyinfra.api import operation
 from pyinfra.facts.choco import ChocoPackages
@@ -9,8 +11,8 @@ from pyinfra.facts.choco import ChocoPackages
 from .util.packaging import ensure_packages
 
 
-@operation
-def packages(packages=None, present=True, latest=False):
+@operation()
+def packages(packages: str | list[str] | None = None, present=True, latest=False):
     """
     Add/remove/update ``choco`` packages.
 
