@@ -2,16 +2,17 @@
 Manage dnf packages and repositories. Note that dnf package names are case-sensitive.
 """
 
+from __future__ import annotations
+
 from pyinfra import host, state
 from pyinfra.api import operation
 from pyinfra.facts.rpm import RpmPackageProvides, RpmPackages
 
-from . import files
 from .util.packaging import ensure_packages, ensure_rpm, ensure_yum_repo
 
 
 @operation(is_idempotent=False)
-def key(src):
+def key(src: str):
     """
     Add dnf gpg keys with ``rpm``.
 
@@ -35,21 +36,21 @@ def key(src):
     yield "rpm --import {0}".format(src)
 
 
-@operation
+@operation()
 def repo(
-    src,
+    src: str,
     present=True,
-    baseurl=None,
-    description=None,
+    baseurl: str | None = None,
+    description: str | None = None,
     enabled=True,
     gpgcheck=True,
-    gpgkey=None,
+    gpgkey: str | None = None,
 ):
     # NOTE: if updating this docstring also update `yum.repo`
     """
     Add/remove/update dnf repositories.
 
-    + name: URL or name for the ``.repo``   file
+    + src: URL or name for the ``.repo``   file
     + present: whether the ``.repo`` file should be present
     + baseurl: the baseurl of the repo (if ``name`` is not a URL)
     + description: optional verbose description
@@ -81,9 +82,7 @@ def repo(
     """
 
     yield from ensure_yum_repo(
-        state,
         host,
-        files,
         src,
         baseurl,
         present,
@@ -94,8 +93,8 @@ def repo(
     )
 
 
-@operation
-def rpm(src, present=True):
+@operation()
+def rpm(src: str, present=True):
     # NOTE: if updating this docstring also update `yum.rpm`
     """
     Add/remove ``.rpm`` file packages.
@@ -118,7 +117,7 @@ def rpm(src, present=True):
         )
     """
 
-    yield from ensure_rpm(state, host, files, src, present, "dnf")
+    yield from ensure_rpm(state, host, src, present, "dnf")
 
 
 @operation(is_idempotent=False)
@@ -130,24 +129,24 @@ def update():
     yield "dnf update -y"
 
 
-_update = update  # noqa: E305 (for use below where update is a kwarg)
+_update = update._inner  # noqa: E305 (for use below where update is a kwarg)
 
 
-@operation
+@operation()
 def packages(
-    packages=None,
+    packages: str | list[str] | None = None,
     present=True,
     latest=False,
     update=False,
     clean=False,
     nobest=False,
-    extra_install_args=None,
-    extra_uninstall_args=None,
+    extra_install_args: str | None = None,
+    extra_uninstall_args: str | None = None,
 ):
     """
     Install/remove/update dnf packages & updates.
 
-    + packages: list of packages to ensure
+    + packages: packages to ensure
     + present: whether the packages should be installed
     + latest: whether to upgrade packages without a specified version
     + update: run ``dnf update`` before installing packages
@@ -207,8 +206,5 @@ def packages(
         upgrade_command="dnf update -y",
         version_join="=",
         latest=latest,
-        expand_package_fact=lambda package: host.get_fact(
-            RpmPackageProvides,
-            name=package,
-        ),
+        expand_package_fact=lambda package: host.get_fact(RpmPackageProvides, package=package),
     )
