@@ -2,6 +2,8 @@
 Manage git repositories and configuration.
 """
 
+from __future__ import annotations
+
 import re
 
 from pyinfra import host
@@ -13,10 +15,8 @@ from . import files, ssh
 from .util.files import chown, unix_path_join
 
 
-@operation(
-    pipeline_facts={"git_config": "repo"},
-)
-def config(key, value, multi_value=False, repo=None):
+@operation()
+def config(key: str, value: str, multi_value=False, repo: str | None = None):
     """
     Manage git config for a repository or globally.
 
@@ -54,27 +54,23 @@ def config(key, value, multi_value=False, repo=None):
 
     if not multi_value and existing_config.get(key) != [value]:
         yield '{0} {1} "{2}"'.format(base_command, key, value)
-        existing_config[key] = [value]
 
     elif multi_value and value not in existing_config.get(key, []):
         yield '{0} --add {1} "{2}"'.format(base_command, key, value)
-        existing_config.setdefault(key, []).append(value)
 
     else:
         host.noop("git config {0} is set to {1}".format(key, value))
 
 
-@operation(
-    pipeline_facts={"git_branch": "target"},
-)
+@operation()
 def repo(
-    src,
-    dest,
-    branch=None,
+    src: str,
+    dest: str,
+    branch: str | None = None,
     pull=True,
     rebase=False,
-    user=None,
-    group=None,
+    user: str | None = None,
+    group: str | None = None,
     ssh_keyscan=False,
     update_submodules=False,
     recursive_submodules=False,
@@ -105,7 +101,7 @@ def repo(
     """
 
     # Ensure our target directory exists
-    yield from files.directory(dest)
+    yield from files.directory._inner(dest)
 
     # Do we need to scan for the remote host key?
     if ssh_keyscan:
@@ -113,7 +109,7 @@ def repo(
         domain = re.match(r"^[a-zA-Z0-9]+@([0-9a-zA-Z\.\-]+)", src)
 
         if domain:
-            yield from ssh.keyscan(domain.group(1))
+            yield from ssh.keyscan._inner(domain.group(1))
         else:
             raise OperationError(
                 "Could not parse domain (to SSH keyscan) from: {0}".format(src),
@@ -130,21 +126,11 @@ def repo(
             git_commands.append("clone {0} --branch {1} .".format(src, branch))
         else:
             git_commands.append("clone {0} .".format(src))
-
-        host.create_fact(GitBranch, kwargs={"repo": dest}, data=branch)
-        host.create_fact(
-            Directory,
-            kwargs={"path": git_dir},
-            data={"user": user, "group": group},
-        )
-
     # Ensuring existing repo
     else:
         if branch and host.get_fact(GitBranch, repo=dest) != branch:
             git_commands.append("fetch")  # fetch to ensure we have the branch locally
             git_commands.append("checkout {0}".format(branch))
-            host.create_fact(GitBranch, kwargs={"repo": dest}, data=branch)
-
         if pull:
             if rebase:
                 git_commands.append("pull --rebase")
@@ -171,19 +157,19 @@ def repo(
 
 @operation()
 def worktree(
-    worktree,
-    repo=None,
+    worktree: str,
+    repo: str | None = None,
     detached=False,
-    new_branch=None,
-    commitish=None,
+    new_branch: str | None = None,
+    commitish: str | None = None,
     pull=True,
     rebase=False,
-    from_remote_branch=None,
+    from_remote_branch: tuple[str, str] | None = None,
     present=True,
     assume_repo_exists=False,
     force=False,
-    user=None,
-    group=None,
+    user: str | None = None,
+    group: str | None = None,
 ):
     """
     Manage git worktrees.
@@ -293,7 +279,6 @@ def worktree(
 
     # Doesn't exist & we want it
     if not host.get_fact(Directory, path=worktree) and present:
-
         # be sure that `repo` is a GIT repository
         if not assume_repo_exists and not host.get_fact(
             Directory,
@@ -321,12 +306,8 @@ def worktree(
         if user or group:
             yield chown(worktree, user, group, recursive=True)
 
-        host.create_fact(Directory, kwargs={"path": worktree}, data={"user": user, "group": group})
-        host.create_fact(GitTrackingBranch, kwargs={"repo": worktree}, data=new_branch)
-
     # It exists and we don't want it
     elif host.get_fact(Directory, path=worktree) and not present:
-
         command = "cd {0} && git worktree remove .".format(worktree)
 
         if force:
@@ -334,12 +315,8 @@ def worktree(
 
         yield command
 
-        host.delete_fact(Directory, kwargs={"path": worktree})
-        host.create_fact(GitTrackingBranch, kwargs={"repo": worktree})
-
     # It exists and we still want it => pull/rebase it
     elif host.get_fact(Directory, path=worktree) and present:
-
         # pull the worktree only if it's already linked to a tracking branch or
         # if a remote branch is set
         if host.get_fact(GitTrackingBranch, repo=worktree) or from_remote_branch:
@@ -359,11 +336,11 @@ def worktree(
             yield command
 
 
-@operation
+@operation()
 def bare_repo(
-    path,
-    user=None,
-    group=None,
+    path: str,
+    user: str | None = None,
+    group: str | None = None,
     present=True,
 ):
     """
@@ -384,7 +361,7 @@ def bare_repo(
         )
     """
 
-    yield from files.directory(path, present=present)
+    yield from files.directory._inner(path, present=present)
 
     if present:
         head_filename = unix_path_join(path, "HEAD")
@@ -397,9 +374,3 @@ def bare_repo(
         else:
             if (user and head_file["user"] != user) or (group and head_file["group"] != group):
                 yield chown(path, user, group, recursive=True)
-
-        host.create_fact(
-            File,
-            kwargs={"path": head_filename},
-            data={"user": user, "group": group, "mode": None},
-        )
