@@ -2,6 +2,8 @@
 Manage OpenRC init services.
 """
 
+from __future__ import annotations
+
 from pyinfra import host
 from pyinfra.api import operation
 from pyinfra.facts.openrc import OpenrcEnabled, OpenrcStatus
@@ -9,14 +11,14 @@ from pyinfra.facts.openrc import OpenrcEnabled, OpenrcStatus
 from .util.service import handle_service_control
 
 
-@operation
+@operation()
 def service(
-    service,
+    service: str,
     running=True,
     restarted=False,
     reloaded=False,
-    command=None,
-    enabled=None,
+    command: str | None = None,
+    enabled: bool | None = None,
     runlevel="default",
 ):
     """
