@@ -1,16 +1,18 @@
+from __future__ import annotations
+
 import re
 from datetime import datetime
 
 from pyinfra.api import QuoteString, StringCommand
 
 
-def unix_path_join(*parts):
-    parts = list(parts)
-    parts[0:-1] = [part.rstrip("/") for part in parts[0:-1]]
-    return "/".join(parts)
+def unix_path_join(*parts) -> str:
+    part_list = list(parts)
+    part_list[0:-1] = [part.rstrip("/") for part in part_list[0:-1]]
+    return "/".join(part_list)
 
 
-def ensure_mode_int(mode):
+def ensure_mode_int(mode: str | int | None) -> int | str | None:
     # Already an int (/None)?
     if isinstance(mode, int) or mode is None:
         return mode
@@ -26,19 +28,19 @@ def ensure_mode_int(mode):
     return mode
 
 
-def get_timestamp():
+def get_timestamp() -> str:
     return datetime.now().strftime("%y%m%d%H%M")
 
 
 def sed_replace(
-    filename,
-    line,
-    replace,
-    flags=None,
+    filename: str,
+    line: str,
+    replace: str,
+    flags: list[str] | None = None,
     backup=False,
     interpolate_variables=False,
-):
-    flags = "".join(flags) if flags else ""
+) -> StringCommand:
+    flags_str = "".join(flags) if flags else ""
 
     line = line.replace("/", r"\/")
     replace = str(replace)
@@ -57,7 +59,7 @@ def sed_replace(
         replace = replace.replace("'", "'\"'\"'")
         sed_script_formatter = "'s/{0}/{1}/{2}'"
 
-    sed_script = sed_script_formatter.format(line, replace, flags)
+    sed_script = sed_script_formatter.format(line, replace, flags_str)
 
     sed_command = StringCommand(
         "sed",
@@ -73,7 +75,7 @@ def sed_replace(
     return sed_command
 
 
-def chmod(target, mode, recursive=False):
+def chmod(target: str, mode: str | int, recursive=False) -> StringCommand:
     args = ["chmod"]
     if recursive:
         args.append("-R")
@@ -83,7 +85,13 @@ def chmod(target, mode, recursive=False):
     return StringCommand(" ".join(args), QuoteString(target))
 
 
-def chown(target, user, group=None, recursive=False, dereference=True):
+def chown(
+    target: str,
+    user: str | None = None,
+    group: str | None = None,
+    recursive=False,
+    dereference=True,
+) -> StringCommand:
     command = "chown"
     user_group = None
 
@@ -107,7 +115,7 @@ def chown(target, user, group=None, recursive=False, dereference=True):
     return StringCommand(" ".join(args), user_group, QuoteString(target))
 
 
-def adjust_regex(line, escape_regex_characters):
+def adjust_regex(line: str, escape_regex_characters: bool) -> str:
     """
     Ensure the regex starts with '^' and ends with '$' and escape regex characters if requested
     """
