@@ -1,4 +1,3 @@
-import os
 import signal
 import sys
 
@@ -15,12 +14,14 @@ pyinfra.is_cli = True
 # Don't write out deploy.pyc/config.pyc etc
 sys.dont_write_bytecode = True
 
+sys.path.append(".")
+
 # Shut it click
 click.disable_unicode_literals_warning = True  # type: ignore
 
 # Force line buffering
-sys.stdout = os.fdopen(sys.stdout.fileno(), "w", 1)
-sys.stderr = os.fdopen(sys.stderr.fileno(), "w", 1)
+sys.stdout.reconfigure(line_buffering=True)  # type: ignore
+sys.stderr.reconfigure(line_buffering=True)  # type: ignore
 
 
 def _handle_interrupt(signum, frame):
