@@ -1,10 +1,11 @@
 from unittest.mock import MagicMock, patch
 
 from pyinfra.api import Config, State
-from pyinfra.api.arguments import get_executor_kwarg_keys, pop_global_arguments
+from pyinfra.api.arguments import CONNECTOR_ARGUMENT_KEYS, pop_global_arguments
 from pyinfra.api.connect import connect_all
 from pyinfra.api.exceptions import PyinfraError
 from pyinfra.api.facts import get_facts
+from pyinfra.connectors.util import CommandOutput, OutputLine
 from pyinfra.facts.server import Arch, Command
 
 from ..paramiko_util import PatchSSHTestCase
@@ -16,7 +17,7 @@ def _get_executor_defaults(state, host):
     return {
         key: value
         for key, value in global_argument_defaults.items()
-        if key in get_executor_kwarg_keys()
+        if key in CONNECTOR_ARGUMENT_KEYS
     }
 
 
@@ -29,19 +30,18 @@ class TestFactsApi(PatchSSHTestCase):
 
         connect_all(state)
 
-        with patch("pyinfra.connectors.ssh.run_shell_command") as fake_run_command:
-            fake_run_command.return_value = MagicMock(), [("stdout", "some-output")]
+        with patch("pyinfra.connectors.ssh.SSHConnector.run_shell_command") as fake_run_command:
+            fake_run_command.return_value = True, CommandOutput(
+                [OutputLine("stdout", "some-output")]
+            )
             fact_data = get_facts(state, Command, ("yes",))
 
         assert fact_data == {anotherhost: "some-output"}
 
         fake_run_command.assert_called_with(
-            state,
-            anotherhost,
             "yes",
             print_input=False,
             print_output=False,
-            return_combined_output=True,
             **_get_executor_defaults(state, anotherhost),
         )
 
@@ -52,31 +52,29 @@ class TestFactsApi(PatchSSHTestCase):
         anotherhost = inventory.get_host("anotherhost")
 
         connect_all(state)
-        anotherhost.current_op_global_kwargs = {
-            "sudo": True,
-            "sudo_user": "someuser",
-            "use_sudo_password": True,
-            "su_user": "someuser",
-            "timeout": 10,
-            "env": {"HELLO": "WORLD"},
+        anotherhost.current_op_global_arguments = {
+            "_sudo": True,
+            "_sudo_user": "someuser",
+            "_su_user": "someuser",
+            "_timeout": 10,
+            "_env": {"HELLO": "WORLD"},
         }
 
-        with patch("pyinfra.connectors.ssh.run_shell_command") as fake_run_command:
-            fake_run_command.return_value = MagicMock(), [("stdout", "some-output")]
+        with patch("pyinfra.connectors.ssh.SSHConnector.run_shell_command") as fake_run_command:
+            fake_run_command.return_value = True, CommandOutput(
+                [OutputLine("stdout", "some-output")]
+            )
             fact_data = get_facts(state, Command, ("yes",))
 
         assert fact_data == {anotherhost: "some-output"}
 
         defaults = _get_executor_defaults(state, anotherhost)
-        defaults.update(anotherhost.current_op_global_kwargs)
+        defaults.update(anotherhost.current_op_global_arguments)
 
         fake_run_command.assert_called_with(
-            state,
-            anotherhost,
             "yes",
             print_input=False,
             print_output=False,
-            return_combined_output=True,
             **defaults,
         )
 
@@ -88,7 +86,7 @@ class TestFactsApi(PatchSSHTestCase):
 
         connect_all(state)
 
-        with patch("pyinfra.connectors.ssh.run_shell_command") as fake_run_command:
+        with patch("pyinfra.connectors.ssh.SSHConnector.run_shell_command") as fake_run_command:
             fake_run_command.return_value = False, MagicMock()
 
             with self.assertRaises(PyinfraError) as context:
@@ -97,12 +95,9 @@ class TestFactsApi(PatchSSHTestCase):
         assert context.exception.args[0] == "No hosts remaining!"
 
         fake_run_command.assert_called_with(
-            state,
-            anotherhost,
             "fail command",
             print_input=False,
             print_output=False,
-            return_combined_output=True,
             **_get_executor_defaults(state, anotherhost),
         )
 
@@ -113,23 +108,21 @@ class TestFactsApi(PatchSSHTestCase):
         anotherhost = inventory.get_host("anotherhost")
 
         connect_all(state)
-        anotherhost.current_op_global_kwargs = {
-            "ignore_errors": True,
+        anotherhost.in_op = True
+        anotherhost.current_op_global_arguments = {
+            "_ignore_errors": True,
         }
 
-        with patch("pyinfra.connectors.ssh.run_shell_command") as fake_run_command:
+        with patch("pyinfra.connectors.ssh.SSHConnector.run_shell_command") as fake_run_command:
             fake_run_command.return_value = False, MagicMock()
             fact_data = get_facts(state, Command, ("fail command",))
 
         assert fact_data == {anotherhost: None}
 
         fake_run_command.assert_called_with(
-            state,
-            anotherhost,
             "fail command",
             print_input=False,
             print_output=False,
-            return_combined_output=True,
             **_get_executor_defaults(state, anotherhost),
         )
 
@@ -141,22 +134,21 @@ class TestFactsApi(PatchSSHTestCase):
 
         connect_all(state)
 
-        with patch("pyinfra.connectors.ssh.run_shell_command") as fake_run_command:
-            fake_run_command.return_value = MagicMock(), [("stdout", "some-output")]
+        with patch("pyinfra.connectors.ssh.SSHConnector.run_shell_command") as fake_run_command:
+            fake_run_command.return_value = MagicMock(), CommandOutput(
+                [OutputLine("stdout", "some-output")]
+            )
             fact_data = get_facts(state, Command, ("yes",), {"_sudo": True})
 
         assert fact_data == {anotherhost: "some-output"}
 
         defaults = _get_executor_defaults(state, anotherhost)
-        defaults["sudo"] = True
+        defaults["_sudo"] = True
 
         fake_run_command.assert_called_with(
-            state,
-            anotherhost,
             "yes",
             print_input=False,
             print_output=False,
-            return_combined_output=True,
             **defaults,
         )
 
@@ -169,22 +161,21 @@ class TestFactsApi(PatchSSHTestCase):
 
         connect_all(state)
 
-        with patch("pyinfra.connectors.ssh.run_shell_command") as fake_run_command:
-            fake_run_command.return_value = MagicMock(), [("stdout", "some-output")]
+        with patch("pyinfra.connectors.ssh.SSHConnector.run_shell_command") as fake_run_command:
+            fake_run_command.return_value = True, CommandOutput(
+                [OutputLine("stdout", "some-output")]
+            )
             fact_data = get_facts(state, Command, ("yes",))
 
         assert fact_data == {anotherhost: "some-output"}
 
         defaults = _get_executor_defaults(state, anotherhost)
-        defaults["sudo"] = True
+        defaults["_sudo"] = True
 
         fake_run_command.assert_called_with(
-            state,
-            anotherhost,
             "yes",
             print_input=False,
             print_output=False,
-            return_combined_output=True,
             **defaults,
         )
 
@@ -197,14 +188,16 @@ class TestFactsApi(PatchSSHTestCase):
         anotherhost.data._sudo_user = "this-should-be-overridden"
         anotherhost.data._su_user = "this-should-be-overridden"
 
-        anotherhost.current_op_global_kwargs = {
-            "su_user": "override-su-user",
+        anotherhost.current_op_global_arguments = {
+            "_su_user": "override-su-user",
         }
 
         connect_all(state)
 
-        with patch("pyinfra.connectors.ssh.run_shell_command") as fake_run_command:
-            fake_run_command.return_value = MagicMock(), [("stdout", "some-output")]
+        with patch("pyinfra.connectors.ssh.SSHConnector.run_shell_command") as fake_run_command:
+            fake_run_command.return_value = True, CommandOutput(
+                [OutputLine("stdout", "some-output")]
+            )
             fact_data = get_facts(
                 state,
                 Command,
@@ -215,44 +208,17 @@ class TestFactsApi(PatchSSHTestCase):
         assert fact_data == {anotherhost: "some-output"}
 
         defaults = _get_executor_defaults(state, anotherhost)
-        defaults["sudo"] = True
-        defaults["sudo_user"] = "override-sudo-user"
-        defaults["su_user"] = "override-su-user"
+        defaults["_sudo"] = True
+        defaults["_sudo_user"] = "override-sudo-user"
+        defaults["_su_user"] = "override-su-user"
 
         fake_run_command.assert_called_with(
-            state,
-            anotherhost,
             "yes",
             print_input=False,
             print_output=False,
-            return_combined_output=True,
             **defaults,
         )
 
-    def test_get_fact_cached(self):
-        inventory = make_inventory(hosts=("anotherhost",))
-        state = State(inventory, Config())
-
-        fact_hash = "a-fact-hash"
-        cached_fact = {"this is a cached fact"}
-        anotherhost = inventory.get_host("anotherhost")
-        anotherhost.facts[fact_hash] = cached_fact
-
-        connect_all(state)
-
-        with patch("pyinfra.connectors.ssh.run_shell_command") as fake_run_command:
-            fake_run_command.return_value = MagicMock(), [("stdout", "some-output")]
-            fact_data = get_facts(
-                state,
-                Command,
-                args=("yes",),
-                kwargs={"_sudo": True},
-                fact_hash=fact_hash,
-            )
-
-        assert fact_data == {anotherhost: cached_fact}
-        fake_run_command.assert_not_called()
-
     def test_get_fact_no_args(self):
         inventory = make_inventory(hosts=("host-1",))
         state = State(inventory, Config())
@@ -262,18 +228,17 @@ class TestFactsApi(PatchSSHTestCase):
         host_1 = inventory.get_host("host-1")
         defaults = _get_executor_defaults(state, host_1)
 
-        with patch("pyinfra.connectors.ssh.run_shell_command") as fake_run_command:
-            fake_run_command.return_value = MagicMock(), [("stdout", "some-output")]
+        with patch("pyinfra.connectors.ssh.SSHConnector.run_shell_command") as fake_run_command:
+            fake_run_command.return_value = MagicMock(), CommandOutput(
+                [OutputLine("stdout", "some-output")]
+            )
             fact_data = get_facts(state, Arch)
 
         assert fact_data == {host_1: "some-output"}
         fake_run_command.assert_called_with(
-            state,
-            host_1,
-            Arch.command,
+            Arch().command(),
             print_input=False,
             print_output=False,
-            return_combined_output=True,
             **defaults,
         )
 
@@ -288,18 +253,17 @@ class TestHostFactsApi(PatchSSHTestCase):
         host_1 = inventory.get_host("host-1")
         defaults = _get_executor_defaults(state, host_1)
 
-        with patch("pyinfra.connectors.ssh.run_shell_command") as fake_run_command:
-            fake_run_command.return_value = MagicMock(), [("stdout", "some-output")]
+        with patch("pyinfra.connectors.ssh.SSHConnector.run_shell_command") as fake_run_command:
+            fake_run_command.return_value = MagicMock(), CommandOutput(
+                [OutputLine("stdout", "some-output")]
+            )
             fact_data = host_1.get_fact(Command, command="echo hello world")
 
         assert fact_data == "some-output"
         fake_run_command.assert_called_with(
-            state,
-            host_1,
             "echo hello world",
             print_input=False,
             print_output=False,
-            return_combined_output=True,
             **defaults,
         )
 
@@ -311,20 +275,19 @@ class TestHostFactsApi(PatchSSHTestCase):
 
         host_1 = inventory.get_host("host-1")
         defaults = _get_executor_defaults(state, host_1)
-        defaults["sudo"] = True
+        defaults["_sudo"] = True
 
-        with patch("pyinfra.connectors.ssh.run_shell_command") as fake_run_command:
-            fake_run_command.return_value = MagicMock(), [("stdout", "some-output")]
+        with patch("pyinfra.connectors.ssh.SSHConnector.run_shell_command") as fake_run_command:
+            fake_run_command.return_value = MagicMock(), CommandOutput(
+                [OutputLine("stdout", "some-output")]
+            )
             fact_data = host_1.get_fact(Command, command="echo hello world", _sudo=True)
 
         assert fact_data == "some-output"
         fake_run_command.assert_called_with(
-            state,
-            host_1,
             "echo hello world",
             print_input=False,
             print_output=False,
-            return_combined_output=True,
             **defaults,
         )
 
@@ -336,19 +299,18 @@ class TestHostFactsApi(PatchSSHTestCase):
 
         host_1 = inventory.get_host("host-1")
         defaults = _get_executor_defaults(state, host_1)
-        defaults["sudo"] = True
+        defaults["_sudo"] = True
 
-        with patch("pyinfra.connectors.ssh.run_shell_command") as fake_run_command:
-            fake_run_command.return_value = MagicMock(), [("stdout", "some-output")]
+        with patch("pyinfra.connectors.ssh.SSHConnector.run_shell_command") as fake_run_command:
+            fake_run_command.return_value = MagicMock(), CommandOutput(
+                [OutputLine("stdout", "some-output")]
+            )
             fact_data = host_1.get_fact(Arch, _sudo=True)
 
         assert fact_data == "some-output"
         fake_run_command.assert_called_with(
-            state,
-            host_1,
-            Arch.command,
+            Arch().command(),
             print_input=False,
             print_output=False,
-            return_combined_output=True,
             **defaults,
         )
