@@ -83,11 +83,11 @@ class TestDockerConnector(TestCase):
         host.connect()
         out = host.run_shell_command(
             command,
-            stdin="hello",
-            get_pty=True,
+            _stdin="hello",
+            _get_pty=True,
             print_output=True,
         )
-        assert len(out) == 3
+        assert len(out) == 2
         assert out[0] is True
 
         command = make_unix_command(command).get_raw_value()
@@ -112,7 +112,7 @@ class TestDockerConnector(TestCase):
 
         host = inventory.get_host("@docker/not-an-image")
         host.connect()
-        out = host.run_shell_command(command, success_exit_codes=[1])
+        out = host.run_shell_command(command, _success_exit_codes=[1])
         assert out[0] is True
 
     def test_run_shell_command_error(self):
